
from ruamel.yaml import CommentedMap, CommentedSeq
from typing import Callable, Any
from re import Pattern
from . import operations
from loguru import logger

ExtractFunction = Callable[[Any, CommentedMap |
                            CommentedSeq, CommentedMap | CommentedSeq], None]
TransformerFunction = Callable[[str, Any, dict[str, Any]], Any]
logger.add("dikdik", format="{time} {level} {message}", level="DEBUG")

def operation_to_transformer(all_operations: dict[Any, Any]=operations.DEFAULT_OPERATIONS, op_name_field: str = "op", op_params_field: str = "params") -> TransformerFunction:
  """
    Mapp the operation name to the operation function and return a TransformerFunction that apply the operation function.
  """
  def generic_transformer(key: str, value: Any, groups: dict[str, Any]) -> Any:
    logger.debug("", key=key, value=value, groups=groups)
    if op_name_field not in groups:
      return None
    if op_params_field not in groups:
      groups['params'] = []

    op = groups[op_name_field]
    params = groups[op_params_field].strip().split(' ')  # split the parameters

    if op in all_operations:
      # Call the specific operation function
      logger.debug("Operation", op=op, params=params, value=value, path=key)
      try:
        trans = all_operations[op](params=params, value=value, path=key)
        return trans
      except TypeError as e:
        raise e
      except operations.OperationVarUndefinedValue as e:
        raise e

  return generic_transformer

def extract_comment_regexp(regexp: Pattern[str], transform: TransformerFunction | None = None) -> ExtractFunction:
  """
  Return an ExtractFunction if a comment match a given regular expression pattern.
  Optionally Apply a set of transformations to the extracted items
  Args:
    regexp (Pattern[str]): The regular expression pattern used to match and extract information from comments.
    transform (TransformerFunction): A function that applies a transformation to the extracted items.
      Transform function signature must be `transform(key: str, value: Any, groups: dict[str,Any]) -> Any`

  Returns:
    ExtractFunction: A function that extracts information from a CommentedMap.
  """

  def extract(key: Any, src: CommentedMap | CommentedSeq, dest: CommentedSeq | CommentedMap) -> None:
      for comment_key, comments in src.ca.items.items():

          # Skip if key does not match
          if key != comment_key:
            continue

          # Skip if empty comments
          if comments is None or len(comments) < 3 or comments[2] is None:
            continue

          groups = {}

          # expand the regexp serching for the operation and parameters
          words = regexp.search(comments[2].value)
          if words is not None and len(words.groups()) > 1:
            # we have a match
            logger.debug("Matched", key=key, value=src[key], groups=words.groupdict())

            # extract the regexp groups as dictionary of groupName: match
            groups.update(words.groupdict())

            # apply the trasformation if any
            if transform is not None:
              ret = transform(key, src[key], groups)
              dest[key] = ret
              if ret is None:
                del dest[key]

  return extract


def deep_visit(src: CommentedMap | CommentedSeq, dest: CommentedMap | CommentedSeq, extract_fun: ExtractFunction) -> None:
  """
    Recursively visits a nested dictionary and apply the given extraction function to each element.

    Args:
      src (CommentedMap | CommentedSeq): The source dictionary to be visited.
      dest (CommentedMap | CommentedSeq): The destination dictionary to be populated, it must be a deep copy of the source.
      extract_fun (Callable): The extraction function to be applied to each element.
      The signature of the function must be `extract_fun(key: Any, src: CommentedMap|CommentedSeq, dest: CommentedMap|CommentedSeq) -> None`
  """

  if isinstance(src, list):
    for key in src:
      index = src.index(key)
      if isinstance(key, dict):
        deep_visit(key, dest[index], extract_fun)
      elif isinstance(key, list):
        deep_visit(key, dest[index], extract_fun)
      extract_fun(key, src, dest)

  if isinstance(src, dict):
    for key, value in src.items():
      if isinstance(src[key], dict):
        deep_visit(src[key], dest[key], extract_fun)
      elif isinstance(src[key], list):
        deep_visit(src[key], dest[key], extract_fun)
      extract_fun(key, src, dest)



#!/usr/bin/env python
#
# This file is part of dik-dik (https://github.com/mbovo/dikdik).
# Copyright (c) 2020-2023 Manuel Bovo.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
from typing import Callable
from ruamel.yaml import YAML, CommentedMap
import sys
from re import Pattern

def dump_yaml(*args):
    yaml = YAML()
    for arg in args:
        if isinstance(arg, tuple):
            data, output = arg
            if output is None or output == "" or output == "-":
                yaml.dump(data, sys.stdout)
                return
            with open(output, 'w') as f:
                yaml.dump(data, f)
        else:
            yaml.dump(arg, sys.stdout)


def extract_regexp(items: dict, regexp: Pattern[str], d: CommentedMap, parent: str) -> list[dict[str, str]]:
  '''
  Extracts informations from CommentedMap on a regular expression pattern.

  Args:
    items (dict): The dictionary containing the comments to process
    regexp (Pattern[str]): The regular expression pattern used to match and extract information from comments.
    d (CommentedMap): The original CommentedMap object containing the comments and values.
    parent (str): The parent key used to construct the path of the extracted information.

  Returns:
    list[dict[str, str]]: A list of dictionaries containing the extracted information.

  '''
  ret = []
  for key, comments in items:
    # Match the comment with the regex and extract the operation and parameters
    if comments is None or len(comments) < 3 or comments[2] is None:
      continue
    groups = regexp.search(comments[2].value)
    if groups is not None and len(groups.groups()) > 1:
      retDict = {
        'path': parent + "." + key,
        'value': d[key]
      }
      retDict.update(groups.groupdict())
      retDict['params'] = retDict['params'].strip().split(
        ' ')  # split the parameters
      ret.append(retDict)
  return ret


def deep_visit(d: CommentedMap, parent: str, regexp: Pattern[str], extract_fun: Callable) -> list[dict[str, str]]:
  """
  Recursively visits a nested dictionary and extracts sub-dictionaries that contain a specific comment.

  Args:
    d (CommentedMap): The nested dictionary to be visited (as returned from ruamel.yaml)
    parent (str): The parent path of the current dictionary (default is an empty string).
    regexp (Pattern): The regular expression to be used to match the comment.
    extract_fun (Callable): The function used to extract the information from the comment. Must have the following signature:
      def extract_fun(items: dict, regexp: Pattern[str], d: CommentedMap, parent: str) -> list[dict[str, str]]:
        pass
  Returns:
    list[dict[str, str]]: A list of dictionaries containing the extracted information.
      Each dictionary contains the following keys:
      - 'path': The path of the current dictionary in the nested structure. eg: secure.admin.password
      - 'value': The value associated with the current nested structure.
      - The dictionary of any regexp match group found in the comment (if any)
  """

  def dv(d: CommentedMap, parent: str = "") -> list[dict[str, str]]:

    ret: list[dict] = []

    items = extract_fun(d.ca.items.items(), regexp, d, parent)
    if items is not None:
      ret = ret + items
    # Recursively visit sub-dictionaries and lists

    for k, v in d.items():
      if isinstance(v, dict):
        inner = dv(v, parent + "." + k)  # type: ignore
        if inner is not None and len(inner) > 0:
          ret = ret + inner
      elif isinstance(v, list):
        for i in v:
          if isinstance(i, dict):
            inner = dv(i, parent + "." + k)  # type: ignore
            if inner is not None and len(inner) > 0:
              ret = ret + inner
    return ret

  return dv(d, parent)



#!/usr/bin/env python
#
# This file is part of dik-dik (https://github.com/mbovo/dikdik).
# Copyright (c) 2020-2023 Manuel Bovo.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
import yaml
import json
import os

from io import TextIOWrapper
from typing import Any, Union, Tuple
from jsonschema import ValidationError, validate

class PowerDict(dict):
    '''
      Powerdict is an advanced dict class
    '''

    def __init__(self, d: dict = {}, **kwargs):
        self._data = {}
        self._data.update(d)
        if kwargs:
            self._data.update(**kwargs)

    def __len__(self) -> int:
        return len(self._data)

    def __getitem__(self, key: str) -> Any:
        if key in self._data:
            return self._data[key]
        raise KeyError(key)

    def __setitem__(self, key, item):
        self._data[key] = item

    def __delitem__(self, key):
        del self._data[key]

    def __iter__(self):
        return iter(self._data)

    def __contains__(self, key: object) -> bool:
        return self._data.__contains__(key)

    def __repr__(self) -> str:
        return repr(self._data)

    def __str__(self) -> str:
        return str(self._data)

    def __eq__(self, value: object) -> bool:
        return self._data.__eq__(value)

    def __copy__(self):
        inst = self.__class__.__new__(self.__class__)
        inst.__dict__.update(self.__dict__)
        # Create a copy and avoid triggering descriptors
        inst.__dict__["data"] = self.__dict__["data"].copy()
        return inst

    def copy(self):
        if self.__class__ is PowerDict:
            return PowerDict(self._data.copy())
        import copy
        data = self._data
        try:
            self._data = {}
            c = copy.copy(self)
        finally:
            self._data = data
        c.update(self)
        return c

    def items(self):
        return self._data.items()

    def pop(self, key, default=None):
        return self._data.pop(key, default)

    def popitem(self):
        return self._data.popitem()

    def clear(self):
        self._data.clear()

    def get(self, key, default=None):
        return self._data.get(key, default)

    def setdefault(self, key, default=None):
        return self._data.setdefault(key, default)

    def keys(self):
        return self._data.keys()

    def values(self):
        return self._data.values()

    def update(self, *args, **kwargs):
        if args:
            if len(args) > 1:
                raise TypeError(
                    "update expected at most 1 arguments, got %d" % len(args))
            other = args[0]  # type: ignore
            if isinstance(other, PowerDict):
                other = other.data  # type: ignore
            elif not isinstance(other, dict):
                other = dict(other)
            self._data.update(other)
        for key, value in kwargs.items():
            self._data[key] = value

    @classmethod
    def fromkeys(cls, iterable, value=None):
        d = cls()
        for key in iterable:
            d[key] = value
        return d

    def validate(self, schema: dict, raise_exception: bool = False) -> Union[None, Tuple[bool, str]]:
        if not raise_exception:
            try:
                validate(self._data, schema)
                return True, ""
            except ValidationError as e:
                return False, e.message
        else:
            validate(self._data, schema)

    def to_yaml(self, file_or_path: Union[str, TextIOWrapper]):
        if isinstance(file_or_path, str):
            with open(file_or_path, 'w') as file:
                yaml.safe_dump(self._data, file)
        else:
            yaml.safe_dump(self._data, file_or_path)

    def from_yaml(self, file_or_path: Union[str, TextIOWrapper], merge: bool = False):
        new_data = {}
        if isinstance(file_or_path, str):
            with open(file_or_path, 'r') as file:
                new_data = yaml.safe_load(file.read())
        else:
            new_data = yaml.safe_load(file_or_path.read())

        if merge:
            new_data = deep_merge(self._data, new_data)
        self.update(new_data)

    def to_json(self, file_or_path: Union[str, TextIOWrapper, None] = None) -> Union[None, str]:
        if file_or_path is None:
            return json.dumps(self._data, indent=2)
        if isinstance(file_or_path, str):
            with open(file_or_path, 'w') as file:
                json.dump(self._data, file)
        else:
            json.dump(self._data, file_or_path)

    def from_json(self, file_or_path: Union[str, TextIOWrapper], merge: bool = False):
        new_data = {}
        if isinstance(file_or_path, str):
            with open(file_or_path, 'r') as file:
                new_data = json.load(file)
        else:
            new_data = json.load(file_or_path)

        if merge:
            new_data = deep_merge(self._data, new_data)
        self.update(new_data)

    def from_env(self, prefix: str = ''):
        for key, value in os.environ.items():
            if key.startswith(prefix):
                nkey = key[len(prefix):]
                self._data[nkey] = value

    def from_dict(self, d: dict, merge: bool = False):
        if merge:
            d = deep_merge(self._data, d)
        self.update(d)

    def to_dict(self) -> dict:
        return self._data

def deep_merge(d1:dict, d2:dict)-> dict:
    """
    Merges two dictionaries recursively.
    """
    merged = d1.copy()
    for k, v in d2.items():
        if isinstance(v, dict) and k in merged and isinstance(merged[k], dict):
            merged[k] = deep_merge(merged[k], v)
        elif isinstance(v, list) and k in merged and isinstance(merged[k], list):
            merged[k].extend(v)
        else:
            merged[k] = v
    return merged


def set_nested(d: dict, keys: list[str], value: str) -> None:
  """
  Sets a nested value in a dictionary based on a list of keys.

  Args:
    d (dict): The dictionary to modify.
    keys (list[str]): A list of keys representing the nested structure.
    value (str): The value to set.

  Returns:
    None
  """
  for key in keys[:-1]:
    d = d.setdefault(key, {})
  if keys[-1] is not None:
    d[keys[-1]] = value

def set_path(d: dict, keys: str, value: str) -> None:
    """
    Sets a nested value in a dictionary based on a string of keys separated by dot.
        eg: set_path(d, 'a.b.c', 1) is equivalent to set_nested(d, ['a', 'b', 'c'], 1) and sets d['a']['b']['c'] = 1
    Args:
        d (dict): The dictionary to modify.
        keys (str): A string of keys representing the nested structure.
        value (str): The value to set.

    Returns:
        None
    """
    set_nested(d, keys.split('.'), value)

def get_nested(d: dict, keys: list[str]) -> Any:
  """
  Gets a nested value in a dictionary based on a list of keys.

  Args:
    d (dict): The dictionary to search.
    keys (list[str]): A list of keys representing the nested structure.

  Returns:
    Any: The value found in the nested structure.
  """
  p = d.copy()
  for key in keys:
    p = p[key]
  return p

def get_path(d: dict, keys: str) -> Any:
  """
  Gets a nested value in a dictionary based on a string of keys separated by dot.
    eg: get_path(d, 'a.b.c') is equivalent to get_nested(d, ['a', 'b', 'c']) and returns d['a']['b']['c']
  Args:
    d (dict): The dictionary to search.
    keys (str): A string of keys representing the nested structure.

  Returns:
    Any: The value found in the nested structure.
  """
  return get_nested(d, keys.split('.'))


import uuid
import random
import string
from typing import Any, Optional, Callable
import jinja2

class OperationError(Exception):

    """Baseclass for all Operation errors."""

    def __init__(self, message: Optional[str] = None) -> None:
        super().__init__(message)

    @property
    def message(self) -> Optional[str]:
        return self.args[0] if self.args else None


class OperationNotFound(OperationError):
   """Raised if an operation does not exist."""


class OperationVarUndefinedValue(OperationError):
    """Raised if the value in the template is undefined"""


def op_keep(**kwargs) -> Any:
  """
    Keep the value as is.
  """
  if 'value' in kwargs:
    return kwargs['value']
  raise OperationError("No value provided for keep() operation")


def op_remove(**kwargs) -> Any:
  """
    Remove the key entirely
  """
  return None


def op_random_uuid(**kwargs) -> Any:
  """
    Return a random UUID
  """
  return str(uuid.uuid1())


def op_random_pasword(**kwargs) -> Any:
    """
    Return a random password

    Args:
      params (list): The length of the password and the characters to use (optional)
    """

    if 'params' not in kwargs:
        raise OperationError(
            "Invalid number of parameters for random_password() operation, expected 1 got 0")

    params = kwargs['params']

    length = int(params[0]) if params[0].isdigit() else 32
    characters = string.ascii_letters + string.digits
    if len(params) > 1:
       characters = params[1]

    password = ''.join(random.choice(characters) for _ in range(length))
    return password


def op_env(**kwargs) -> Any:
  """
    Return an environment variable or a default value

    Args:
      params (list): The name of the environment variable
      value (str): The default value to return if the environment variable is not set
  """
  if 'params' not in kwargs:
    raise OperationError(
        "Invalid number of parameters for env() operation, expected 1 got 0")

  if 'value' not in kwargs:
    raise OperationError(
        "Invalid number of parameters for env() operation, expected 1 got 0")

  params = kwargs['params']
  value = kwargs['value']

  import os
  return os.getenv(params[0], value)


def op_value(**kwargs) -> Any:
  """
    Change the value with the one given in the parameters

    Args:
      params (list): The new value to set
  """
  if 'params' in kwargs:
    params = kwargs['params']
    if len(params) == 0:
      raise OperationError("No parameters provided for value() operation")
    elif len(params) == 1:
      return params[0]
    else:
      return ", ".join(params)
  raise OperationError

def generate_op_var(cfg: dict, root_key:str ='vars') -> Callable:
  """
    Configure the var() operation with passed configuration dictionary and an optional root key
  """

  def op_var(**kwargs) -> Any:
    """
      Return the value of a variable defined in the configuration

      Args:
        params (list): The name of the variable to return
    """
    if 'params' not in kwargs:
      raise OperationError(
          "Invalid number of parameters for var() operation, expected 1 got 0")

    params = kwargs['params']

    if len(params) != 1:
      raise OperationError(
          f"Invalid number of parameters for var() operation, expected 1 got {len(params)}")

    p = params[0]
    try:
      tpl = jinja2.Template(f"{{{{ {p} }}}}")
      if 'vars' not in cfg:
        return ""
      return tpl.render(cfg['vars'])
    except jinja2.exceptions.UndefinedError as e:
      raise OperationVarUndefinedValue(str(e))
    except KeyError as e:
      raise OperationVarUndefinedValue(str(e))
    except Exception as e:
      raise OperationError(str(e))

  return op_var


DEFAULT_OPERATIONS = {
    'keep': op_keep,
    'remove': op_remove,
    'random_uuid': op_random_uuid,
    'random_password': op_random_pasword,
    'env': op_env,
}